;;;
;;;  $Revision: 1.1 $ of $Date: 1998/05/07 23:57:07 $
;;;
;;;  Copyright (C) 1994, 1995, 1998. Doss and Narasimhan
;;;
;;;  Hani J. Doss (doss@stat.ohio-state.edu) and 
;;;  B. Narasimhan (naras@stat.stanford.edu)
;;; 
;;;  This program is free software; you can redistribute it and/or modify
;;;  it under the terms of the GNU General Public License as published by
;;;  the Free Software Foundation; either version 2 of the License, or
;;;  (at your option) any later version.
;;; 
;;;  This program is distributed in the hope that it will be useful,
;;;  but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;  GNU General Public License for more details.
;;; 
;;;  You should have received a copy of the GNU General Public License
;;;  along with this program; if not, write to the Free Software
;;;  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;; 
(defpackage "UTILITY" (:use "XLISP"))
(in-package "UTILITY")
(import '(user::combine))
(defmeth text-item-proto :width (&optional width)
  (if width
      (let ((sz (slot-value 'size)))
        (setf (slot-value 'size) (list width (select sz 1))))
    (select (slot-value 'size) 0)))
(defmeth button-item-proto :width (&optional width)
  (if width
      (let ((sz (slot-value 'size)))
        (setf (slot-value 'size) (list width (select sz 1))))
    (select (slot-value 'size) 0)))
(defmeth interval-scroll-item-proto :width (&optional width)
  (if width
      (let ((sz (slot-value 'size)))
        (setf (slot-value 'size) (list width (select sz 1))))
    (select (slot-value 'size) 0)))
(defmeth interval-scroll-item-proto :value (&optional (val nil set))
  (let ((interval (slot-value 'xlisp::interval))
        (num-points (slot-value 'xlisp::num-points)))
    (if set 
        (let* ((min (elt interval 0))
               (max (elt interval 1))
               (val (if (= min max)
                      0
                   (floor (* (1- num-points) (/ (- val min) (- max min)))))))
          (call-next-method val)
          (send self :display-value)
          (send self :user-action)))
    (let ((min (elt interval 0))
          (max (elt interval 1)))
      (+ min (* (/ (call-next-method) (max 1 (1- num-points))) 
         (- max min))))))
(defproto value-item-proto '(format)
  () text-item-proto
  "A dialog item for formatted displaying of values of slider stops.")
(export 'value-item-proto)
(defmeth value-item-proto :isnew (string &rest args &key (format "~g"))
  "Method args: (string &rest args &key (format ~g))
Creates an instance of value-item-proto; format is used for formatted
printing. It can be a format string or a list of two numbers. In the
latter case, variable format printing is assumed using ~v,vf."
  (setf (slot-value 'format) format)
  (apply #'call-next-method string args))
(defmeth value-item-proto :format (&optional format)
  "Method args: (&optional format)
Sets or retrieves the format, which should be a format string or a
list of 2 numbers for ~v,~vf printing."
  (if format
      (setf (slot-value 'format) format)
    (slot-value 'format)))
(defmeth interval-scroll-item-proto :display-value ()
  (when (slot-value 'xlisp::value-text-item)
        (let* ((v-item (slot-value 'xlisp::value-text-item))
               (format (send v-item :format))
               (value (send self :value)))
          (if (listp format)
              (send v-item :text 
                    (format nil "~v,vf" 
                            (select format 0) (select format 1) value))
            (send v-item :text (format nil format value))))))
(defmeth interval-scroll-item-proto :value-item ()
  (slot-value 'xlisp::value-text-item))
(defun arrange (list &key rows cols)
  "Method args: (list &key rows cols)
Chops up a list into sublists of rows items each and
returns a list of lists. If rows is not given, it is 
floor of sqrt of length of list. Rows need not divide list length!"
  (let ((n (length list)))
    (cond
     (rows
      (unless (>= n rows)
              (error "number of items less than number of rows"))
      (multiple-value-bind (q rem) (floor n rows)
        (if (= rem 0)
            (mapcar #'(lambda(x) (select list (+ x (* rows (iseq q)))))
                    (iseq rows))
          (append  
           (mapcar #'(lambda(x) (select list (+ x (* rows (iseq (1+ q))))))
                   (iseq rem))
           (mapcar #'(lambda(x) (select list (+ x (* rows (iseq q)))))
                   (iseq rem (1- rows)))))))
     (t
      (unless cols (setf cols (floor (sqrt n))))
      (unless (>= n cols)
        (error "number of items less than number of cols"))
      (let* ((q (floor n cols))
             (product (* cols q)))
        (if (= product n)
            (split-list list cols)
          (append  
           (split-list (select list (iseq product)) cols)
           (list (select list (+ product (iseq (- n product))))))))))))

(defun ok-or-abort-dialog (string)
  "Args: string.
Prompts the user with a string in a dialog."
  (let ((text (send text-item-proto :new string))
        (ok (send modal-button-proto :new "OK"))
        (abort (send modal-button-proto :new "Abort"
                     :action #'top-level)))
    (send (send modal-dialog-proto :new (list text (list ok abort)))
          :modal-dialog)))
(defun get-a-string-dialog (s &key initial)
  "Args: (s &key initial)
Opens a modal dialog with prompt S, a text field and OK and Abort buttons.
INITIAL is converted to a string with ~A format directive. Returns
string of text field content on OK, returns to top-level on Abort."
  (let* ((text (send text-item-proto :new s))
         (edit-item (send edit-text-item-proto :new
                          (if initial (format nil "~a" initial) "")
                          :text-length 40))
         (abort (send modal-button-proto :new "Abort"
                      :action #'top-level))
         (ok (send modal-button-proto :new "OK" 
                   :action #'(lambda() (send edit-item :text)))))
    (send 
     (send modal-dialog-proto :new (list text edit-item (list ok abort)))
     :modal-dialog)))
(defun get-nonempty-string-dialog (s &key initial)
  "Args: (s &key initial)
Opens a modal dialog with prompt S, a text field and OK button.
INITIAL is converted to a string with ~A format directive. Returns
string of text field content on OK. Will return a nonempty string or
take user back to top-level."
(loop
 (let ((input (get-a-string-dialog s :initial initial)))
   (if (= (length input) 0)
       (message-dialog
        "Error!!\n\nNeed a non-empty string!")
     (return input)))))
(defun get-a-value-dialog (s &key (initial "" suppliedp))
  "Args: (s &key (initial \"\" suppliedp)
Opens a modal dialog with prompt S, a text field and OK button.
INITIAL is converted to a string with ~S format directive. Returns
string of text field content on OK."
  (let* ((initial (if suppliedp
                       (format nil "~s" initial)))
         (result (get-a-string-dialog s :initial initial)))
    (if result (list (read (make-string-input-stream result) nil)))))
(defun get-nonnil-value-dialog (s &key (initial "" suppliedp))
  "Args: (s &key initial)
Opens a modal dialog with prompt S, a text field and OK button.
INITIAL is converted to a string with ~A format directive. Returns
string of text field content on OK. Will return a nonnil value or
take user back to top-level."
(loop
 (let* ((input (if suppliedp
                   (get-a-value-dialog s :initial initial)
                 (get-a-value-dialog s))))
   (if (some #'not input)
       (message-dialog
        "Error!!\n\nNeed a non-nil value!")
     (return input)))))
(defun get-tested-value-dialog (s &key (initial "" suppliedp) 
                                     (test #'(lambda(x) t))
                                     (error-message "Please retry"))
  "Args: (s &key initial (test #'(lambda(x) t)) 
                         (error-message \"Please Retry\"))
Opens a modal dialog with prompt S, a text field and OK button.
INITIAL is converted to a string with ~A format directive. Returns
string of text field content on OK. Will return a value or
take user back to top-level. "
(loop
 (let* ((input (if suppliedp
                   (get-a-value-dialog s :initial initial)
                 (get-a-value-dialog s))))
   (if (some #'not (mapcar test input))
       (message-dialog error-message)
     (return input)))))
(defun get-yes-no-list (info-string list &key (default nil))
  "Args: (info-string list &key (default nil))
Prompts the user for a Yes/No answer in a dialog box. Info-string
is a informative string. List is a list of strings that go against
the check-boxes. Default is all boxes not checked."
  (let* ((info (send text-item-proto :new info-string))
         (check-items 
          (if (and default (listp default))
              (mapcar #'(lambda(x) 
                          (send toggle-item-proto :new x :value y))
                      list default)
            (mapcar #'(lambda(x) 
                        (send toggle-item-proto :new x :value
                              default))
                    list)))
         (ok (send modal-button-proto :new "OK"
                   :action
                   #'(lambda() 
                       (mapcar #'(lambda(x) (send x :value)) check-items))))
         (d (send modal-dialog-proto :new 
                  (list (list info) (arrange check-items) (list ok)))))
    (send d :modal-dialog)))
(defun convert-to-numbers (list-of-strings)
  "Args: list-of-strings
Converts a list of strings into numeric values. Returns nil if stymied."
  (let ((values 
         (ignore-errors (mapcar #'read-from-string list-of-strings))))
    (if (notevery #'numberp values)
        nil
      values)))
(defun some-files-dont-exist (list-of-file-names)
  "Args: list-of-files
Returns true unless all files exist."
  (some #'not (mapcar #'probe-file list-of-file-names)))
(defun coerce-to-layout-of (y x)
  "Args: (x y)
Returns a list of contents of x, laid out like y. Y is merely
referenced and should be either a list or a list of lists."
  (let* ((items x)
         (n (length items))
         (m (length (combine y)))
         (reversed-result ())
         (count 0))
    (unless (= m n)
            (error "list lengths are unequal."))
    (dolist (element y)
            (let ((l (if (listp element)
                         (length element)
                       1)))
              (cond 
               ((listp element) 
                (setf reversed-result 
                      (cons 
                       (select items (+ count (iseq l))) reversed-result)))
               (t 
                (setf reversed-result 
                      (cons (select items count) reversed-result))))
              (setf count (+ count l))))
    (reverse reversed-result)))
(defun strcat (&rest args)
  "Args: (&rest args)
Concatenates a bunch of strings and returns the result."
  (apply #'concatenate (append (list 'string) args)))
(defun definedp (symbol)
  "Method args: symbol
Returns t if symbol is defined and bound."
  (and (boundp symbol) symbol))
(defun ask-user-for-value (&rest args)
  "Method args: &rest args
Asks the user to enter value in a dialog with a prompt string which
is all of args concatenated."
  (get-value-dialog (apply #'strcat args)))
(defun make-sliders (triples &key layout formats value-text-lengths 
                             no-of-slider-stops)
  "Args: (triples &key layout formats value-text-lengths slider-stops)
Triples is a list consisting of triples: a label, an interval, and an
action function taking a single argument. Layout is used if
provided. Format is a list of formats suitable for value-item-proto.
Value-text-lengths defaults to a list of 10's if formats is not
specified or a list and no-of-slider-stops to a list of 51's."
  (let ((n (length triples))
        (slider-items ())
        (sliders-alone ()))
    (when (not formats)
          (setf formats (repeat "~g" n)))
    (when (not value-text-lengths)
          (setf value-text-lengths (repeat 10 n)))
    (when (not no-of-slider-stops)
          (setf no-of-slider-stops (repeat 51 n)))
    (flet ((make-item (a b c d)
                      (let* ((ti (send text-item-proto :new (select a 0)))
                             (vi (send value-item-proto :new ""
                                       :format b
                                       :text-length 
                                       (if (listp b)
                                           (select b 0)
                                         c)))
                             (si (send interval-scroll-item-proto :new
                                       (select a 1)
                                       :points d
                                       :text-item vi
                                       :action (select a 2))))
                        (send si :width (+ (send ti :width)
                                           (send vi :width)))
                        (if layout 
                            (list (list ti vi) si)
                          (list (list (list ti vi) si))))))
          (setf slider-items (mapcar #'make-item triples formats
                               value-text-lengths no-of-slider-stops)))
    (setf sliders-alone
          (if layout
              (map-elements #'select slider-items 1)
            (mapcar #'(lambda(x) (select (select x 0) 1))
                    slider-items)))
    (when layout
          (setf slider-items (coerce-to-layout-of layout slider-items)))
    (values slider-items sliders-alone)))
(export
        '(value-item-proto
          arrange
          ok-or-abort-dialog
          get-a-string-dialog
          get-nonempty-string-dialog
          get-a-value-dialog
          get-nonnil-value-dialog
          get-tested-value-dialog
          get-yes-no-list
          convert-to-numbers
          some-files-dont-exist
          coerce-to-layout-of
          strcat
          definedp
          ask-user-for-value
          modified-boxplot
          make-sliders))
(provide "utility")
